import collections
from glob import glob

import vi3o
from vi3o import mjpg
from vi3o import utils
from vi3o import compat

_VideoBlock = collections.namedtuple(
    "_VideoBlock", ["video", "timestamp_offset", "systime_offset"]
)


class VideoCat(object):
    """
    Concatenates multiple video files into a single video object. The *videos* parameter
    is a list of videos to be concatenated. It can either be a list of filenames or a list
    of other *Video* objects. Typical usage:

    .. code-block:: python

        from vi3o import VideoCat

        for img in VideoCat(['part1.mkv', 'part2.mkv']):
            ...

    """

    def __init__(self, videos, **kwargs):
        self._videos = []
        self._length = 0

        for values in videos:
            timestamp_offset, systime_offset = None, None

            try:
                vid, timestamp_offset, systime_offset = values
            except ValueError:
                vid = values

            if isinstance(vid, compat.basestring):
                vid = vi3o.Video(vid, **kwargs)

            if isinstance(vid, mjpg.Mjpg) and (
                timestamp_offset is not None or systime_offset is not None
            ):
                # TODO: Provide an implementation for mjpg movies that has the same
                #       behaviour w.r.t systime and timestamp as the Mkv implementation
                raise NotImplementedError()

            self._length += len(vid)
            self._videos.append(_VideoBlock(vid, timestamp_offset, systime_offset))

        self._systimes = []  # type: List[float]
        for blk in self._videos:
            if blk.systime_offset is None:
                self._systimes.extend(blk.video.systimes)
            else:
                self._systimes.extend(
                    (
                        blk_frame[0] / 1.0e6 + blk.systime_offset
                        for blk_frame in blk.video.frame
                    )
                )

    def __len__(self):
        return self._length

    def __iter__(self):
        frame_index = 0
        for blk in self._videos:
            for frame in blk.video:
                if blk.systime_offset is not None:
                    frame.systime = frame.timestamp + blk.systime_offset

                if blk.timestamp_offset is not None:
                    frame.timestamp += blk.timestamp_offset

                frame.index = frame_index
                frame_index += 1
                yield frame

    def __getitem__(self, item):
        if isinstance(item, slice):
            return utils.SlicedView(self, item, {"systimes": self._sliced_systimes})

        if isinstance(item, int):
            # Support negative indices
            if item < 0:
                item += self._length

            # Quick exit for excessive ranges
            if not 0 <= item < self._length:
                raise IndexError

            # Keep track of the requested frame index
            frame_index = item

            # Find the correct block
            for blk in self._videos:
                length = len(blk.video)
                # Break if the index lies in [previous_length, blk.cumulative_length)
                if item < length:
                    break
                item -= length

            else:
                # Unreachable
                raise RuntimeError("Oups")  # pragma: no cover

            # Collect the frame with the correct relative index
            frame = blk.video[item]  # pylint: disable=undefined-loop-variable

            # Update the systime of the frame
            if blk.systime_offset is not None:
                frame.systime = (
                    frame.timestamp
                    + blk.systime_offset  # pylint: disable=undefined-loop-variable
                )
            if blk.timestamp_offset is not None:
                frame.timestamp += blk.timestamp_offset

            frame.index = frame_index
            return frame

        raise TypeError

    @property
    def systimes(self):
        return self._systimes

    def _sliced_systimes(self, range_):
        return [self._systimes[i] for i in range_]

    @property
    def videos(self):
        """ Compatibility method """
        return [blk.video for blk in self._videos]


class VideoGlob(VideoCat):
    """
    Subclass of :class:`VideoCat` that is initiated with a :py:func:`glob.glob` wildcard string instead of
    a list of videos. The wildcard is expanded into a list of filenames that is the sorted before
    concatenated.
    """

    def __init__(self, pathname):
        videos = glob(pathname)
        videos.sort()
        VideoCat.__init__(self, videos)



import sys
import datetime

if sys.version_info >= (3, 3, 0):
    import pathlib

    basestring = (str, bytes)

    def utc_datetime_to_epoch(dt):  # pylint: disable=invalid-name
        # Note that the parsed datetime object must be "aware" for datetime.timestamp()
        # to produce the expected result. This means that we must set the timezone to
        # UTC explicitly.
        return dt.replace(tzinfo=datetime.timezone.utc).timestamp()

    FileNotFoundError = FileNotFoundError
else:
    import pathlib2 as pathlib

    basestring = basestring

    def utc_datetime_to_epoch(dt):  # pylint: disable=invalid-name
        # Keep datetimes naive to avoid any inadvertent time zone conversions
        return (
            dt.replace(tzinfo=None) - datetime.datetime(1970, 1, 1)
        ).total_seconds() + dt.microsecond / 1.0e6

    class FileNotFoundError(OSError):
        pass



import cv2, os
import numpy as np

windows = {}

def _on_mouse(event, x, y, flags, name):
    if event == cv2.EVENT_LBUTTONDOWN:
        print "Click:", x, y

def view(img, name="Video"):
    if name not in windows:
        windows[name] = cv2.namedWindow(name)
        cv2.setMouseCallback(name, _on_mouse, name)
    if img.dtype != 'B':
        img = np.minimum(np.maximum(img, 0), 255).astype('B')
    cv2.imshow(name, img)
    while True:
        key = cv2.waitKey(1) & 0xFF
        if key in (ord('q'), 27):
            exit(0)
        elif key == ord(' '):
            view.paused = not view.paused
        elif key == ord('\n'):
            return
        if not view.paused:
            return

view.paused = False




from vi3o import Video
import os

mydir = os.path.dirname(__file__)
test_mjpg = os.path.join(mydir, "t.mjpg")
systime_mkv = os.path.join(mydir, "systime.mkv")


def test_video():
    assert Video(test_mjpg)[1].systime == 1445859308.97
    assert Video(systime_mkv)[1].systime == 1448984844.2525



from vi3o.image import *
from test.util import TempDir
import numpy as np
from py.test import raises
from vi3o.compat import pathlib

mydir = os.path.dirname(__file__)
test_jpg = os.path.join(mydir, "00000000.jpg")
test_jpg2 = os.path.join(mydir, "img00012680.jpg")


def test_save_load():
    with TempDir() as d:
        img = np.zeros((320, 240, 3), 'B')
        img[10, 20] = 255
        imsave(img, os.path.join(d, "t1.jpg"), "jpg")
        imsave(img, os.path.join(d, "t2.png"))
        im1 = imread(os.path.join(d, "t1.jpg"))
        assert all(im1[10, 20] > 250)
        im2 = imread(os.path.join(d, "t2.png"))
        assert all(im2[10, 20] == 255)

        fimg = np.zeros((320, 240, 3))
        fimg[10, 20] = 0.1
        fimg[30, 40] = -0.1
        imsavesc(fimg, os.path.join(d, "t3.png"))
        im3 = imread(os.path.join(d, "t3.png"))
        assert all(im3[10, 20] == 255)
        assert all(im3[30, 40] == 0)
        assert all(im3[1, 2] == 127)

def test_scale():
    img = np.zeros((240, 320, 3), 'B')
    assert imscale(img, 2).shape == (480, 640, 3)
    assert imscale(img, (640, 480)).shape == (480, 640, 3)

    img = np.zeros((240, 320), 'B')
    assert imscale(img, 0.5).shape == (120, 160)
    assert imscale(img, (160, 120)).shape == (120, 160)

def test_imsave_and_imread_pathlib():
    with TempDir() as d:
        img = np.zeros((320, 240, 3), 'B')
        img[10, 20] = 255
        path = pathlib.Path("t1.jpg")
        imsave(img, path, "jpg")
        imread(path)


def test_imgdir():
    with TempDir() as d:
        img = np.zeros((320, 240, 3), 'B')
        out = ImageDirOut(d)
        out.view(img)
        out.view(img)
        out.view(img)
        assert len(os.listdir(d)) == 3

        out = ImageDirOut(d)
        out.view(img)
        out.view(img)
        assert len(os.listdir(d)) == 2

        out = ImageDirOut(d, append=True)
        out.view(img)
        out.view(img)
        assert len(os.listdir(d)) == 4

def test_imrotate():
    img = np.ones([12, 16], 'B') * 255
    # FIXME: Improve test by adding:
    # img[1, :] = img[10, :] = np.array(range(16)) * 10
    # img[:, 1] = img[:, 14] = np.array(range(12)) * 10
    img[0, :] = img[11, :] = 0
    img[:, 0] = img[:, 15] = 0

    rot = imrotate(img, 1.5708/2, [16, 0], [16*2, 12*2], NEAREST)
    # imwrite(rot, "/tmp/t.png")
    # os.system("xzgv /tmp/t.png")

    # pkt = imrotate(img, 1.5708, [16, 0], [16*2, 12*2], point=[16, 0])
    # assert all(np.round(pkt, 6) == [16, 12])

    cnt1 = cnt2 = 0
    for y in range(-12, 24):
        for x in range(-16, 32):
            rx, ry = imrotate(img, 1.5708/2, [16, 0], [16 * 2, 12 * 2], NEAREST, point=[x, y])
            rx, ry = int(round(rx)), int(round(ry))
            if 0 < rx < 16*2 and 0 < ry < 12*2:
                if 0 <= x < 16 and 0 <= y < 12:
                    assert rot[ry, rx] == img[y, x]
                    cnt1 += 1
                else:
                    assert rot[ry, rx] < 10
                    cnt2 += 1
    assert cnt1 > 10
    assert cnt2 > 10

    assert all(rot[:12].flat == 0)

def test_imread():
    img = imread(test_jpg)
    assert isinstance(img, np.ndarray)
    assert img.shape == (288, 360)
    with raises(IOError):
        img = imread(test_jpg2)
    img = imread(test_jpg2, True)
    assert isinstance(img, np.ndarray)
    assert img.shape == (240, 320, 3)


"""
:mod:`vi3o.image` --- Image handling
====================================
"""

import PIL.Image
import numpy as np
import os
from vi3o import view

NEAREST = PIL.Image.NEAREST
BILINEAR = PIL.Image.BILINEAR
BICUBIC = PIL.Image.BICUBIC
ANTIALIAS = PIL.Image.ANTIALIAS


def imsave(img, filename, format=None):
    """
    Save the image *img* into a file named *filename*. If the fileformat is not specified
    in *format*, the filename extension will be used as *format*.
    """
    # Be compatible with pathlib.Path filenames
    filename = str(filename)
    if format is None:
        format = filename.split('.')[-1]
    if format == 'jpg':
        format = 'jpeg'
    if img.dtype != 'B':
        img = np.minimum(np.maximum(img, 0), 255).astype('B')
    PIL.Image.fromarray(img).save(filename, format.lower())

def imsavesc(img, filename, format=None):
    """
    Rescales the intensities of the image *img* to cover the 0..255 range and then calls
    :func:`vi3o.image.imsave` to save it.
    """
    imsave(ptpscale(img), filename, format)

class Silent: pass

def imread(filename, repair=False):
    """
    Load an image from the file *filename*. If *repair* is True, attempts will be made to decode
    broken frames, in which case partially decoded frames might be returned. A warning is printed
    to standard output unless *repair* is set to :class:`vi3o.image.Silent`.
    """
    # Be compatible with pathlib.Path filenames
    a =  PIL.Image.open(str(filename))
    try:
        a.load()
    except IOError as e:
        if not repair:
            raise
        if repair is not Silent:
            print("Warning: IOError while reading '%s': %s" % (filename, e))
    return np.array(a)


def imscale(img, size, interpolation=NEAREST):
    """
    Scales the image *img* into a new size specified by *size*. It can either be a number
    specifying a factor that relates the old size of the new or a 2-tuple with new size as
    *(width, height)*.
    """
    if isinstance(size, float) or isinstance(size, int):
        size = (size * img.shape[1], size * img.shape[0])
    size = map(int, size)
    return np.array(PIL.Image.fromarray(img).resize(size, interpolation))

def imrotate(img, angle, center=None, size=None, interpolation=NEAREST, point=None):
    return imrotate_and_scale(img, angle, 1.0, center, size, interpolation, point)

def imrotate_and_scale(img, angle, scale, center=None, size=None, interpolation=NEAREST, point=None):
    """
    Rotate the image, *img*, *angle* radians around the point *center* which defaults to
    the center of the image. The output image size is specified in *size* as *(width, height)*.
    If *point* is specifed as *(x, y)* it will be taken as a coordinate in the original image
    and be transformed into the corresponing coordinate in te output image and returned
    instead of the image. Integer coordinates are considered to be at the centers of the
    pixels.
    """
    if center is None:
        h, w = img.shape[:2]
        center = (w/2, h/2)
    if size is None:
        h, w = img.shape[:2]
        size = (w, h)
    else:
        size = tuple(size)
    cx, cy = size[0] / 2, size[1] / 2
    s, c = np.sin(angle),  np.cos(angle)
    data = [c, -s, -c*cx + s*cy + center[0],
            s,  c, -s*cx - c*cy + center[1]]
    data = [scale*d for d in data]
    data[2] += (1-scale) * center[0]
    data[5] += (1-scale) * center[1]
    if point is not None:
        a, b, c, d, e, f, _, _, _ = np.linalg.inv(np.vstack((np.reshape(data, (2, 3)), [0, 0, 1]))).flat
        x, y = point
        x += 0.5
        y += 0.5
        return a*x + b*y + c - 0.5, d*x + e*y + f - 0.5
    return np.array(PIL.Image.fromarray(img).transform(size, PIL.Image.AFFINE, data, interpolation))

def ptpscale(img):
    """
    Rescales (and translates) the intensities of the image *img* to cover the 0..255 range.
    """
    a, b = min(img.flat), max(img.flat)
    if a == b:
        return np.zeros_like(img)
    return (img - a) * (255.0 / (b - a))


def imshow(img):
    """
    Display the image *img* in the DebugViewer and pause the viewer with the image showing.
    """
    view(img, pause=True)

def imshowsc(img):
    """
    Rescales (and translates) the intensities of the image *img* to cover the 0..255 range.
    Then display the image *img* in the DebugViewer and pause the viewer with the image showing.
    """
    view(img, scale=True, pause=True)

class ImageDirOut(object):
    """
    Creates a directory called *dirname* for storing a sequence of images in the format
    specified by *format*. If *append* is not *True* (default) the content of the directory
    will be cleard if it excists.
    """
    def __init__(self, dirname, format='jpg', append=False):
        self.dirname = dirname
        self.format = format
        self.fcnt = 0
        if not os.path.exists(dirname):
            os.mkdir(dirname)
        if append:
            self.fcnt = len(os.listdir(dirname))
        else:
            for fn in os.listdir(dirname):
                os.unlink(os.path.join(dirname, fn))

    def view(self, img):
        """
        Save the image *img* as *dirname/xxxxxxxx.format* where *xxxxxxxx* is an 8 digit
        sequence number.
        """
        imsave(img, os.path.join(self.dirname, '%.8d.%s' % (self.fcnt, self.format)), self.format)
        self.fcnt += 1

    def viewsc(self, img):
        """
        Rescales the intensities of the image *img* to cover the 0..255 range, and call
        :py:meth:`~vi3o.image.ImageDirOut.view` to save it.
        """
        self.view(ptpscale(img))

imview = imshow
imviewsc = imshowsc
imload = imread
imwrite = imsave



# Py2 compat
from __future__ import unicode_literals, print_function

import os
import sys

import pytest
import vi3o
from vi3o import cat
from test.util import _FakeVideo, itertools, mock



_LENGTHS = [5, 7, 3]
_BASE_OFFSET = 15

# Calculate offsets for blocks
_TIMESTAMP_OFFSETS = [0] + list(itertools.accumulate(_LENGTHS))
_SYSTIME_OFFSETS = [_BASE_OFFSET + x for x in _TIMESTAMP_OFFSETS]


# Systimes should be sequential with an offset
_EXPECTED_SYSTIMES = [_BASE_OFFSET + x for x in range(sum(_LENGTHS))]

# Timestamps should be sequential with 0 offset
_EXPECTED_TIMESTAMPS = [x for x in range(sum(_LENGTHS))]



@pytest.fixture(scope="function")
def mocked_mkv(monkeypatch):
    # Patch vi3o.mkv with Video object that returns
    # each of _Fake(l0), _Fake(l1), ..., _Fake(ln)
    monkeypatch.setattr(
        vi3o.mkv,
        "Mkv",
        mock.Mock(spec=vi3o.mkv.Mkv, side_effect=[_FakeVideo(l) for l in _LENGTHS]),
    )


@pytest.fixture(scope="function")
def videos(mocked_mkv):
    return [
        (vi3o.Video("hello.mkv"), timestamp_offset, systime_offset)
        for _, timestamp_offset, systime_offset in zip(
            _LENGTHS, _TIMESTAMP_OFFSETS, _SYSTIME_OFFSETS
        )
    ]


@pytest.fixture()
def videocat(videos):
    return cat.VideoCat(videos)


def test_videos_fixture(videos):
    for (video, _, _), expected_length in zip(videos, _LENGTHS):
        assert len(video) == expected_length
    assert len(videos) == len(_LENGTHS)


mydir = os.path.dirname(__file__)
test_mkvs = [os.path.join(mydir, f) for f in ["a.mkv", "b.mkv", "c.mkv"]]


def test_cat():
    all_videos = [vi3o.Video(v) for v in test_mkvs]
    frames = sum(len(v) for v in all_videos)
    video = cat.VideoCat(test_mkvs)

    systimes = []
    for fcnt, img in enumerate(video):
        assert img.index == fcnt
        systimes.append(img.systime)
    assert fcnt == frames - 1

    assert len(video) == frames
    assert video[3].systime == all_videos[0][3].systime
    assert video[-3].systime == all_videos[-1][-3].systime

    index = len(all_videos[0])
    cut = video[index - 2 : index + 2]
    assert len(cut) == 4
    assert cut[0].systime == all_videos[0][-2].systime
    assert cut[3].systime == all_videos[1][1].systime

    assert cut[0].pts == all_videos[0][-2].pts
    assert cut[3].pts == all_videos[1][1].pts == 60000

    assert video.systimes == systimes
    assert cut.systimes == systimes[index - 2 : index + 2]


def test_cat_iterator_raises_on_exhausted(videocat):
    iterator = iter(videocat)
    for length in _LENGTHS:
        for _ in range(length):
            next(iterator)

    # Now the iterator should be exhausted
    with pytest.raises(StopIteration):
        next(iterator)


def test_cat_iterator_correct_times(videocat):
    actual_systimes = []
    actual_timestamps = []
    for f in videocat:
        actual_systimes.append(f.systime)
        actual_timestamps.append(f.timestamp)

    assert actual_systimes == pytest.approx(_EXPECTED_SYSTIMES)
    assert actual_timestamps == pytest.approx(_EXPECTED_TIMESTAMPS)


def test_cat_correct_systimes(videocat):
    assert videocat.systimes == pytest.approx(_EXPECTED_SYSTIMES)


def test_cat_correct_length(videocat):
    assert len(videocat) == sum(_LENGTHS)


def test_cat_getitem_raises_typeerror(videocat):
    with pytest.raises(TypeError):
        _ = videocat["invalid value"]


def test_cat_getitem_backward_skipping(videocat):
    # Skip to block 2
    frame = videocat[sum(_LENGTHS[:2])]
    assert frame.systime != pytest.approx(_BASE_OFFSET)
    # Skip back to block 0
    frame = videocat[0]
    assert frame.systime == pytest.approx(_BASE_OFFSET)
    assert frame.timestamp == pytest.approx(0)


def test_cat_getitem_raises_indexerror_on_missing(videocat):
    with pytest.raises(IndexError):
        _ = videocat[sum(_LENGTHS)]

    with pytest.raises(IndexError):
        _ = videocat[-sum(_LENGTHS) - 1]


def test_cat_getitem_and_iterator_correspond(videocat):
    reference = list(videocat)
    for idx, expected in zip(range(0, sum(_LENGTHS)), reference):
        assert videocat[idx].systime == pytest.approx(expected.systime)
        assert videocat[idx].timestamp == pytest.approx(expected.timestamp)


def test_cat_getitem_negative_indices(videocat):
    length = sum(_LENGTHS)

    actual_systimes = list(videocat[idx].systime for idx in range(-length, 0))
    actual_timestamps = list(videocat[idx].timestamp for idx in range(-length, 0))
    assert actual_systimes == pytest.approx(_EXPECTED_SYSTIMES)
    assert actual_timestamps == pytest.approx(_EXPECTED_TIMESTAMPS)


def test_cat_getitem_slice_systimes(videocat):
    my_slice = slice(_LENGTHS[1] - 2, _LENGTHS[2])
    sliced_values = videocat[my_slice]

    assert sliced_values.systimes == pytest.approx(_EXPECTED_SYSTIMES[my_slice])


def test_cat_getitem_slice_values(videocat):
    my_slice = slice(_LENGTHS[1] - 2, _LENGTHS[2])

    actual_systimes = []
    actual_timestamps = []
    for f in videocat[my_slice]:
        actual_systimes.append(f.systime)
        actual_timestamps.append(f.timestamp)

    assert actual_systimes == pytest.approx(_EXPECTED_SYSTIMES[my_slice])
    assert actual_timestamps == pytest.approx(_EXPECTED_TIMESTAMPS[my_slice])


def test_cat_raises_on_mjpg():
    # With offsets it should raise
    video = mock.MagicMock(spec=vi3o.mjpg.Mjpg)
    with pytest.raises(NotImplementedError):
        _ = cat.VideoCat([(video, 0, 0)])

    # But not without
    _ = cat.VideoCat([video])


def test_cat_kwargs_forwarding(monkeypatch):
    my_mock = mock.MagicMock()
    monkeypatch.setattr(vi3o, "Video", my_mock)
    videocat = cat.VideoCat(["hello.mkv"], grey=True)
    my_mock.assert_called()
    my_mock.assert_called_with("hello.mkv", grey=True)



""" Test the block stitching logic of Recording """
from __future__ import unicode_literals, print_function

import itertools
import sys

import pytest
import vi3o.compat as compat
from vi3o import recording

from test.util import _FakeVideo, itertools, mock

# pylint: disable=missing-docstring
# pylint: disable=redefined-outer-name

_RECORDING_DATA = (
    '<Recording RecordingToken="20190812_113000_B27F_00408C18823A" >'
    "<RecordingGroup> </RecordingGroup>"
    "<SourceToken>4</SourceToken>"
    "<StartTime>2019-08-12T09:30:00.813188Z</StartTime>"
    "<StopTime>2019-08-12T09:45:00.945062Z</StopTime>"
    "<Content></Content>"
    '<Track TrackToken="Video">'
    "<VideoAttributes>"
    "<Width>1920</Width>"
    "<Height>2160</Height>"
    "<Framerate>25.00000</Framerate>"
    "<Framerate_fraction>25:1</Framerate_fraction>"
    "<Encoding>video/x-h264</Encoding>"
    "<Bitrate>0</Bitrate>"
    "</VideoAttributes>"
    "</Track>"
    "<Application>AxisCamera</Application>"
    "<CustomAttributes>"
    "<TriggerTrigger>Record video</TriggerTrigger>"
    "<TriggerName>Record</TriggerName>"
    "<TriggerType>triggered</TriggerType>"
    "</CustomAttributes>"
    "</Recording>"
)

_BLOCK_DATA = (
    '<RecordingBlock RecordingBlockToken="20190808_141501_32D4" >'
    "<RecordingToken>20190808_141501_1778_00408C18823D</RecordingToken>"
    "<StartTime>2019-08-08T12:15:01.338584Z</StartTime>"
    "<StopTime>2019-08-08T12:20:01.704260Z</StopTime>"
    "<Status>Complete</Status>"
    "</RecordingBlock>"
)




# Note that normally only one type of block exists in a recording
_BLOCK_XML_FN_A = "derp.xml"
_BLOCK_MKV_FN_A = "derp.mkv"

_BLOCK_XML_FN_B = "herp.xml"
_BLOCK_MKV_FN_B = "herp.mkv"

_LENGTHS = [5, 7, 3]
_BASE_OFFSET = 15

# Calculate offsets for blocks
_BLOCK_OFFSETS = [
    _BASE_OFFSET + val for val in ([0] + list(itertools.accumulate(_LENGTHS)))
]


# Systimes should be sequential with an offset
_EXPECTED_SYSTIMES = [_BASE_OFFSET + x for x in range(sum(_LENGTHS))]

# Timestamps should be sequential with 0 offset
_EXPECTED_TIMESTAMPS = [x for x in range(sum(_LENGTHS))]


@pytest.fixture(scope="function", autouse=True)
def patched_vi3o(monkeypatch):
    # Patch vi3o.mkv with Video object that returns
    # each of _Fake(l0), _Fake(l1), ..., _Fake(ln)
    monkeypatch.setattr(
        recording.vi3o.mkv,
        "Mkv",
        mock.Mock(side_effect=[_FakeVideo(l) for l in _LENGTHS]),
    )


@pytest.fixture()
def recording_xml_data(tmp_path):
    # Generate a believable example with the following layout:
    #
    # tmp_path/
    # |--folder/
    #    |-- _BLOCK_XML_FN_A
    #        _BLOCK_MKV_FN_A
    #  --folder2/
    #    |-- _BLOCK_XML_FN_B
    #     -- _BLOCK_MKV_FN_B
    #  -- recording.xml
    #
    # and return the filepaths
    recording_xml = tmp_path / "recording.xml"
    recording_xml.write_text(_RECORDING_DATA)

    folder_a = tmp_path / "folder"
    folder_a.mkdir()
    xml_a = folder_a / _BLOCK_XML_FN_A
    video_a = folder_a / _BLOCK_MKV_FN_A
    xml_a.write_text(_BLOCK_DATA)
    video_a.write_bytes(b"")

    folder_b = tmp_path / "folder2"
    folder_b.mkdir()
    xml_b = folder_b / _BLOCK_XML_FN_B
    video_b = folder_b / _BLOCK_MKV_FN_B
    xml_b.write_text(_BLOCK_DATA)
    video_b.write_bytes(b"")
    yield recording_xml, xml_a, video_a, xml_b, video_b


def test_read_recording_xml_parsing(recording_xml_data):
    # Test both pathlib and str interface
    recording_xml, _, _, _, _ = recording_xml_data
    for path in [recording_xml, str(recording_xml)]:
        metadata = recording.read_recording_xml(path)
        # Correct channel parsed
        assert metadata.channel == 4  # SourceToken
        # Correct number of blocks discovered
        assert len(metadata.blocks) == 2
        # Filename of block is the same as the XML
        assert _BLOCK_MKV_FN_A in str(metadata.blocks[0].filename)
        # Alphanumerically ordered blocks
        assert _BLOCK_MKV_FN_B in str(metadata.blocks[1].filename)


@pytest.fixture
def metadata_missing_blocks():
    return recording.RecordingMetadata(
        recording_id="id",
        channel=0,
        start=0.0,
        stop=0.0,
        width=0,
        height=0,
        framerate=25,
        blocks=[],
    )


@pytest.fixture
def mocked_metadata(tmp_path):
    filename = tmp_path / "some_filename.mkv"
    filename.write_bytes(b"")

    blocks = [
        recording.RecordingBlock(
            start=offset, stop=offset + length, filename=filename, status="Complete"
        )
        for offset, length in zip(_BLOCK_OFFSETS, _LENGTHS)
    ]
    yield recording.RecordingMetadata(
        recording_id="id",
        channel=0,
        start=_BASE_OFFSET,
        # The end of the video is the last of the block_offsets
        stop=_BLOCK_OFFSETS[-1],
        width=0,
        height=0,
        framerate=1,
        blocks=blocks,
    )


def test_read_recoding_xml_raises_on_missing_recording_xml(tmp_path):
    with pytest.raises(compat.FileNotFoundError):
        _ = recording.read_recording_xml(tmp_path)


def test_read_recording_xml_raises_on_invalid_block(recording_xml_data):
    recording_xml, xml_a, _, _, _ = recording_xml_data
    xml_a.write_bytes(b"")
    with pytest.raises(RuntimeError):
        recording.read_recording_xml(recording_xml)


def test_read_recording_xml_raises_on_missing_video_for_block(recording_xml_data):
    recording_xml, _, video_a, _, _ = recording_xml_data
    video_a.unlink()
    with pytest.raises(compat.FileNotFoundError):
        _ = recording.read_recording_xml(recording_xml)


def test_read_recording_xml_raises_on_incomplete_status(recording_xml_data):
    recording_xml, xml_a, _, _, _ = recording_xml_data

    xml_a.write_text(_BLOCK_DATA.replace("Complete", "Incomplete"))
    with pytest.raises(RuntimeError):
        _ = recording.read_recording_xml(recording_xml)


def test_recording_raises_on_mjpg(recording_xml_data):
    recording_xml, _, video_a, _, _ = recording_xml_data

    # Note: Below is necessary due to pathlib.Path.replace only being available in Python 3.3+
    video_a.unlink()
    new_video = video_a.parent / (str(video_a.stem) + ".mjpg")
    new_video.write_bytes(b"")
    with pytest.raises(NotImplementedError):
        _ = recording.Recording(recording.read_recording_xml(recording_xml))


def test_recording_raises_on_empty_blocks(metadata_missing_blocks):
    with pytest.raises(RuntimeError):
        _ = recording.Recording(metadata_missing_blocks)


def test_recording_iterator_correct_times(mocked_metadata):
    actual_systimes = []
    actual_timestamps = []
    for f in recording.Recording(mocked_metadata):
        actual_systimes.append(f.systime)
        actual_timestamps.append(f.timestamp)

    assert actual_systimes == pytest.approx(_EXPECTED_SYSTIMES)
    assert actual_timestamps == pytest.approx(_EXPECTED_TIMESTAMPS)


def test_recording_forwards_kwargs(monkeypatch, mocked_metadata):
    my_mock = mock.MagicMock()

    monkeypatch.setattr(recording.cat, "VideoCat", my_mock)

    _ = recording.Recording(mocked_metadata, grey=True)
    my_mock.assert_called()
    _, kwargs = my_mock.call_args
    assert "grey" in kwargs



from setuptools import setup
from setuptools.command.test import test as TestCommand
import sys
import os

# Read the version but avoid importing the __init__.py since
# we might not have all dependencies installed
with open(os.path.join(os.path.dirname(__file__), "vi3o", "version.py")) as fp:
    exec(fp.read())

requirements=["cffi>=1.0.0", "numpy>=1.7.1,<1.17"]
if sys.version_info <= (3, 3, 0):
    requirements.append("pathlib2")

class PyTestCommand(TestCommand):
    user_options = []

    def finalize_options(self):
        pass

    def run_tests(self):
        import pytest
        errno = pytest.main()
        sys.exit(errno)

setup(
    name='vi3o',
    description='VIdeo and Image IO',
    long_description='''
Utility for loading/saving/displaying video and images. It gives random
access to mjpg (in http multipart format) and H264 (in .mkv format) video
frames. For recordings origination from Axis cameras the camera system
time at the time of capture is provided as timestamp for each frame.
    ''',
    version=__version__,
    packages=['vi3o'],
    zip_safe=False,
    url='http://vi3o.readthedocs.org',
    author='Hakan Ardo',
    author_email='hakan@debian.org',
    license='MIT',
    setup_requires=["cffi>=1.0.0"],
    cffi_modules=["build_mjpg.py:ffi", "build_mkv.py:ffi"],
    install_requires=requirements,
    extras_require={
        "full": [
            "pillow < 7",
            "pyglet < 1.5"
        ]
    },
    cmdclass={'test': PyTestCommand},
    tests_require=['pytest <= 4.6, !=4.6.0', 'pillow < 7', "mock"],
)
